/* gpg.cpp
 *
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusLeaf.
 *
 * JanusLeaf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "stdafx.h"
#include "resource.h"
#include "ioutil.h"
#include "gpg.h"
#include "InputBox.h"

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

gpg::gpg(LPCSTR gpgexe, LPCSTR confdir, std::vector<std::string> &secring, std::vector<std::string> &pubring)
{
  this->gpgexe = gpgexe;
  this->confdir = confdir;
  this->secring = secring;
  this->pubring = pubring;
  hProcess = INVALID_HANDLE_VALUE;
  hStdIn = INVALID_HANDLE_VALUE;
  hStdOut = INVALID_HANDLE_VALUE;
  hStdErr = INVALID_HANDLE_VALUE;
}

gpg::~gpg()
{
  _ASSERT(hProcess == INVALID_HANDLE_VALUE);
  if (hProcess != INVALID_HANDLE_VALUE) ::CloseHandle(hProcess);
}

HRESULT gpg::setStdErr(HANDLE hStdErr)
{
  this->hStdErr = hStdErr;
  return S_OK;
}

HRESULT gpg::execute(std::string &cmdline, LPCSTR passphrase, HANDLE hStatusC, HANDLE hCommandC)
{
  HRESULT hr = S_OK;
  HANDLE hPassphraseC = INVALID_HANDLE_VALUE, hPassphraseP = INVALID_HANDLE_VALUE, hNull = INVALID_HANDLE_VALUE;
  PROCESS_INFORMATION pi;
  std::string base_commandline;
  LPSTR commandline2 = NULL;
  
  ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
  hProcess = INVALID_HANDLE_VALUE;
  
  base_commandline = "\"" + gpgexe + "\"";
  base_commandline += " --no-random-seed-file --no-default-keyring --always-trust --utf8-strings --cipher-algo AES256";
  base_commandline += " --homedir=\"" + confdir + "\"";
  if (!pubring.empty()) {
    for (std::vector<std::string>::iterator itr = pubring.begin(); itr != pubring.end(); ++itr) {
      base_commandline += " --keyring=\"" + *itr + "\"";
    }
  }
  if (!secring.empty()) {
    for (std::vector<std::string>::iterator itr = secring.begin(); itr != secring.end(); ++itr) {
      base_commandline += " --secret-keyring=\"" + *itr + "\"";
    }
  }
  if (passphrase) {
    if (FAILED(hr = createWritePipe(&hPassphraseP, &hPassphraseC))) {
      goto finish_func;
    }
    char buff[32 + 1];
    if (FAILED(hr = StringCbPrintf(buff, 32, " --passphrase-fd=%d", reinterpret_cast<int>(hPassphraseC)))) {
      return hr;
    }
    base_commandline += buff;
  }
  if (hCommandC != INVALID_HANDLE_VALUE) {
    char buff[32 + 1];
    if (FAILED(hr = StringCbPrintf(buff, 32, " --command-fd=%d", reinterpret_cast<int>(hCommandC)))) {
      return hr;
    }
    base_commandline += buff;
  }
  if (hStatusC != INVALID_HANDLE_VALUE) {
    char buff[32 + 1];
    if (FAILED(hr = StringCbPrintf(buff, 32, " --status-fd=%d", reinterpret_cast<int>(hStatusC)))) {
      return hr;
    }
    base_commandline += buff;
  }

  base_commandline += cmdline;
  
  SECURITY_ATTRIBUTES sa;
  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = TRUE;
  sa.lpSecurityDescriptor = NULL;

  hNull = CreateFile("nul", GENERIC_READ | GENERIC_WRITE,
                     FILE_SHARE_READ | FILE_SHARE_WRITE,
                     &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hNull == INVALID_HANDLE_VALUE) {
    hr = HRESULT_FROM_WIN32(::GetLastError());
    goto finish_func;
  }

  STARTUPINFO si;
  ZeroMemory(&si, sizeof(STARTUPINFO));
  si.cb = sizeof(STARTUPINFO);
  si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
  si.wShowWindow = SW_HIDE;
  si.hStdInput = (hStdIn == INVALID_HANDLE_VALUE) ? hNull : hStdIn;
  si.hStdOutput = (hStdOut == INVALID_HANDLE_VALUE) ? hNull : hStdOut;
  si.hStdError = (hStdErr == INVALID_HANDLE_VALUE) ? hNull : hStdErr;
  
  commandline2 = new CHAR[base_commandline.size() + 2];
  if (FAILED(hr = StringCbCopy(commandline2, base_commandline.size() + 1, base_commandline.c_str()))) {
    goto finish_func;
  }
  if (!::CreateProcess(NULL, commandline2, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi)) {
    hr = HRESULT_FROM_WIN32(::GetLastError());
    goto finish_func;
  }
  hProcess = pi.hProcess;
  ::CloseHandle(pi.hThread); pi.hThread = NULL;
  delete [] commandline2;
  commandline2 = NULL;
  // pXt[Yo
  if (hPassphraseC != INVALID_HANDLE_VALUE) {
    ::CloseHandle(hPassphraseC);
    hPassphraseC = INVALID_HANDLE_VALUE;

    if (FAILED(hr = write(hPassphraseP, passphrase))) {
      DWORD dwErr = ::GetLastError();
      goto finish_func;
    }
    ::CloseHandle(hPassphraseP);
    hPassphraseP = INVALID_HANDLE_VALUE;
  }
finish_func:
  if (commandline2) delete [] commandline2;
  if (FAILED(hr) && (hProcess != INVALID_HANDLE_VALUE)) ::CloseHandle(hProcess);
  if (hNull != INVALID_HANDLE_VALUE) ::CloseHandle(hNull);
  if (hStdIn != INVALID_HANDLE_VALUE) {
    ::CloseHandle(hStdIn);
    hStdIn = INVALID_HANDLE_VALUE;
  }
  if (hStdOut != INVALID_HANDLE_VALUE) {
    ::CloseHandle(hStdOut);
    hStdOut = INVALID_HANDLE_VALUE;
  }
  if (hStdErr != INVALID_HANDLE_VALUE) {
    ::CloseHandle(hStdErr);
    hStdErr = INVALID_HANDLE_VALUE;
  }
  if (hCommandC != INVALID_HANDLE_VALUE) ::CloseHandle(hCommandC);
  if (hStatusC != INVALID_HANDLE_VALUE) ::CloseHandle(hStatusC);
  if (hPassphraseC != INVALID_HANDLE_VALUE) ::CloseHandle(hPassphraseC);
  if (hPassphraseP != INVALID_HANDLE_VALUE) ::CloseHandle(hPassphraseP);
  return hr;
}

HRESULT gpg::close(LPDWORD lpdwExitCode)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;;

  _ASSERT(hProcess != INVALID_HANDLE_VALUE);

  if (::WaitForSingleObject(hProcess, INFINITE) == WAIT_FAILED) {
    hr = HRESULT_FROM_WIN32(::GetLastError());
    goto finish_func;
  }
  if (!::GetExitCodeProcess(hProcess, &dwExitCode)) {
    hr = HRESULT_FROM_WIN32(::GetLastError());
    goto finish_func;
  }
  if (hProcess != INVALID_HANDLE_VALUE) ::CloseHandle(hProcess);
  hProcess = INVALID_HANDLE_VALUE;
  if (lpdwExitCode) *lpdwExitCode = dwExitCode;
finish_func:
  return S_OK;
}

//////////////////////////////////////////////////////////////////////
// utility
//////////////////////////////////////////////////////////////////////

HRESULT gpg::validateName(LPCSTR name)
{
  LPCSTR p;
  HRESULT hr = S_OK;

  if (::lstrlen(name) < 5) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  if ((*name >= '0') && (*name <= '9')) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  for (p = name; *p; ++p) {
    if ((*p == '<') || (*p == '>')) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
  }
finish_func:
  return hr;
}

HRESULT gpg::validateEMail(LPCSTR email)
{
  DWORD dwAt;
  LPCSTR p;
  HRESULT hr = S_OK;

  if ((*email == '@') ||
      (*(email + ::strlen(email) - 1) == '@') ||
      (*(email + ::strlen(email) - 1) == '.') ||
      (::strstr(email, ".."))) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  for (dwAt = 0, p = email; *p; ++p) {
    if (*p == '@') dwAt++;
  }
  if (dwAt != 1) {
    hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
    goto finish_func;
  }
  for (dwAt = 0, p = email; *p; ++p) {
    if (*p & 0x80) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
    if (*p == '@') {
      dwAt++;
    } else {
      if ((dwAt == 0) &&
          (!::isalnum(*p)) &&
          (*p != '+') &&
          (*p != '-') &&
          (*p != '_') &&
          (*p != '.')) {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        goto finish_func;
      } else if ((dwAt == 1)  &&
                 (!::isalnum(*p)) &&
                 (*p != '+') &&
                 (*p != '-') &&
                 (*p != '_') &&
                 (*p != '.')) {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        goto finish_func;
      }
    }
  }
finish_func:
  return hr;
}

HRESULT gpg::validateComment(LPCSTR comment)
{
  LPCSTR p;
  HRESULT hr = S_OK;

  for (p = comment; *p; ++p) {
    if ((*p == '(') || (*p == ')')) {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      goto finish_func;
    }
  }
finish_func:
  return hr;
}

//////////////////////////////////////////////////////////////////////
// gpg commands
//////////////////////////////////////////////////////////////////////

HRESULT gpg::encryptfile(std::vector<std::string> &recipients, LPCSTR srcfile, LPCSTR dstfile, BOOL bSign, LPCSTR passphrase, LPDWORD lpdwExitCode)
{
  HRESULT hr = S_OK;
  std::string commandline;
  DWORD dwExitCode = 0;

  for (std::vector<std::string>::iterator itr = recipients.begin(); itr != recipients.end(); ++itr) {
    commandline += " -r \"" + (*itr) + "\"";
  }
  if (bSign) {
    commandline += " --sign";
  }
  commandline += " --batch --yes";
  commandline += " -o \"" + std::string(dstfile) + "\"";
  commandline += " --encrypt \"" + std::string(srcfile) + "\"";
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (lpdwExitCode) *lpdwExitCode = dwExitCode;
finish_func:
  return (hr);
}

HRESULT gpg::encryptfile_sym(LPCSTR srcfile, LPCSTR dstfile, BOOL bSign, LPCSTR passphrase, LPCSTR passphrase_sym, LPDWORD lpdwExitCode)
{
  HRESULT hr = S_OK;
  std::string commandline;
  DWORD dwExitCode = 0;
  HANDLE hStatusP = INVALID_HANDLE_VALUE, hStatusC = INVALID_HANDLE_VALUE;
  HANDLE hCommandP = INVALID_HANDLE_VALUE, hCommandC = INVALID_HANDLE_VALUE;

  _ASSERT(passphrase_sym);
  if (!passphrase_sym) return E_UNEXPECTED;
  if (bSign) {
    commandline += " --sign";
  }
// ꍇɃpXt[Y2KvɂȂ邽߁A--batch͎gȂ
//  commandline += " --batch --yes";
  commandline += " -o \"" + std::string(dstfile) + "\"";
  commandline += " --symmetric \"" + std::string(srcfile) + "\"";
  if (FAILED(hr = createWritePipe(&hCommandP, &hCommandC))) goto finish_func;
  if (FAILED(hr = createReadPipe(&hStatusP, &hStatusC))) goto finish_func;
  if (FAILED(hr = execute(commandline, passphrase_sym, hStatusC, hCommandC))) goto finish_func;
  hStatusC = INVALID_HANDLE_VALUE;
  hCommandC = INVALID_HANDLE_VALUE;
  /* gpg X^[gAbv */
  {
    BOOL sym = FALSE;
    std::string str, strstatus, str1;
    DWORD dwRead;
    while (SUCCEEDED(hr = read(hStatusP, str, &dwRead))) {
      if (dwRead == 0) break;
      if (str.substr(0, 9) == "[GNUPG:] ") {
        std::string::size_type pos = str.find(' ', 9);
        if (pos == std::string::npos) {
          strstatus = str.substr(9);
        } else {
          strstatus = str.substr(9, pos - 9);
        }
      } else {
        strstatus = "";
      }
      if ((strstatus == "BEGIN_ENCRYPTION") ||
          (strstatus == "BEGIN_SIGNING")) {
        break;
      } else if (strstatus == "NEED_PASSPHRASE") {
        sym = FALSE;
      } else if (strstatus == "NEED_PASSPHRASE_SYM") {
        sym = TRUE;
      } else if (strstatus == "GET_LINE") {
      } else if (strstatus == "GET_HIDDEN") {
        if (sym) {
          write(hCommandP, passphrase_sym);
          write(hCommandP, "\n");
        } else {
          write(hCommandP, passphrase);
          write(hCommandP, "\n");
        }
      } else if (strstatus == "GET_BOOL") {
        write(hCommandP, "y\n");
      } else if (strstatus == "BAD_PASSPHRASE") {
      } else {
      }
    }
  }
  if (FAILED(hr)) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (lpdwExitCode) *lpdwExitCode = dwExitCode;
finish_func:
  if (hStatusP != INVALID_HANDLE_VALUE) ::CloseHandle(hStatusP);
  if (hStatusC != INVALID_HANDLE_VALUE) ::CloseHandle(hStatusC);
  if (hCommandP != INVALID_HANDLE_VALUE) ::CloseHandle(hCommandP);
  if (hCommandC != INVALID_HANDLE_VALUE) ::CloseHandle(hCommandC);
  return (hr);
}

HRESULT gpg::encrypt_open(std::vector<std::string> &recipients, HANDLE hStdIn, HANDLE hStdOut, BOOL bSign, LPCSTR passphrase, HANDLE hStatusC, HANDLE hCommandC)
{
  HRESULT hr = S_OK;
  _ASSERT(hStdIn != INVALID_HANDLE_VALUE);
  _ASSERT(hStdOut != INVALID_HANDLE_VALUE);
  std::string commandline;
  for (std::vector<std::string>::iterator itr = recipients.begin(); itr != recipients.end(); ++itr) {
    commandline += " -r \"" + (*itr) + "\"";
  }
  if (bSign) {
    commandline += " --sign";
  }
  commandline += " --encrypt";
  this->hStdIn = hStdIn;
  this->hStdOut = hStdOut;
  hr = execute(commandline, passphrase, hStatusC, hCommandC);
  return (hr);
}

HRESULT gpg::encrypt_sym_open(HANDLE hStdIn, HANDLE hStdOut, BOOL bSign, LPCSTR passphrase, HANDLE hStatusC, HANDLE hCommandC)
{
  HRESULT hr = S_OK;
  _ASSERT(hStdIn != INVALID_HANDLE_VALUE);
  _ASSERT(hStdOut != INVALID_HANDLE_VALUE);
  _ASSERT(passphrase);
  std::string commandline;
  if (bSign) {
    commandline += " --sign";
  }
  commandline += " --symmetric";
  this->hStdIn = hStdIn;
  this->hStdOut = hStdOut;
  hr = execute(commandline, passphrase, hStatusC, hCommandC);
  return (hr);
}

HRESULT gpg::decryptfile(LPCSTR srcfile, LPCSTR dstfile, BOOL bVerify, LPCSTR passphrase, LPDWORD lpdwExitCode)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  std::string commandline;
  if (!bVerify) {
    commandline += " --skip-verify";
  }
  commandline += " -o \"";
  commandline += dstfile;
  commandline += "\"";
  commandline += " --batch --yes";
  commandline += " --decrypt";
  commandline += " \"";
  commandline += srcfile;
  commandline += "\"";
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (lpdwExitCode) *lpdwExitCode = dwExitCode;
finish_func:
  return (hr);
}

HRESULT gpg::decrypt_open(HANDLE hStdIn, HANDLE hStdOut, BOOL bVerify, LPCSTR passphrase, HANDLE hStatusC, HANDLE hCommandC)
{
  HRESULT hr = S_OK;

  _ASSERT(hStdIn != INVALID_HANDLE_VALUE);
  _ASSERT(hStdOut != INVALID_HANDLE_VALUE);

  std::string commandline;
  if (!bVerify) {
    commandline += " --skip-verify";
  }
  commandline += " --decrypt";
  this->hStdIn = hStdIn;
  this->hStdOut = hStdOut;
  hr = execute(commandline, passphrase, hStatusC, hCommandC);
  return (hr);
}

HRESULT gpg::sign_detach(LPCSTR srcfile, LPCSTR signfile, LPCSTR passphrase)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  std::string commandline;
  commandline += " -o \"";
  commandline += signfile;
  commandline += "\"";
  commandline += " --detach-sign";
  commandline += " \"";
  commandline += srcfile;
  commandline += "\"";
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (dwExitCode == 0) {
    hr = S_OK;
  } else {
    hr = E_FAIL;
  }
finish_func:
  return (hr);
}

HRESULT gpg::verify(LPCSTR srcfile, LPCSTR signfile, LPCSTR passphrase)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  std::string commandline;
  commandline += " --verify";
  commandline += " \"";
  commandline += signfile;
  commandline += "\"";
  commandline += " \"";
  commandline += srcfile;
  commandline + "\"";
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (dwExitCode == 0) {
    hr = S_OK;
  } else {
    hr = E_FAIL;
  }
finish_func:
  return (hr);
}

HRESULT gpg::genkey(LPCSTR secring, LPCSTR pubring, LPCSTR name, LPCSTR email, LPCSTR comment, LPCSTR passphrase, HANDLE hStatusC, HANDLE hCommandC)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  HANDLE hWrite = INVALID_HANDLE_VALUE;

  _ASSERT(secring != NULL);
  _ASSERT(pubring != NULL);

  std::string commandline;
  commandline += " --batch --gen-key";
  if (FAILED(hr = createWritePipe(&hWrite, &hStdIn))) {
    goto finish_func;
  }
  if (FAILED(hr = execute(commandline, passphrase, hStatusC, hCommandC))) goto finish_func;

  if (FAILED(hr = write(hWrite, "Key-Type: DSA\r\n"))) goto finish_func;
  if (FAILED(hr = write(hWrite, "Key-Length: 1024\r\n"))) goto finish_func;
  if (FAILED(hr = write(hWrite, "Subkey-Type: ELG-E\r\n"))) goto finish_func;
  if (FAILED(hr = write(hWrite, "Subkey-Length: 2048\r\n"))) goto finish_func;
  if (name) {
    if (FAILED(hr = write(hWrite, "Name-Real: "))) goto finish_func;
    if (FAILED(hr = write(hWrite, name))) goto finish_func;
    if (FAILED(hr = write(hWrite, "\r\n"))) goto finish_func;
  }
  if (email) {
    if (FAILED(hr = write(hWrite, "Name-Email: "))) goto finish_func;
    if (FAILED(hr = write(hWrite, email))) goto finish_func;
    if (FAILED(hr = write(hWrite, "\r\n"))) goto finish_func;
  }
  if (comment) {
    if (FAILED(hr = write(hWrite, "Name-Comment: "))) goto finish_func;
    if (FAILED(hr = write(hWrite, comment))) goto finish_func;
    if (FAILED(hr = write(hWrite, "\r\n"))) goto finish_func;
  }
  if (FAILED(hr = write(hWrite, "Expire-Date: 0\r\n"))) goto finish_func;
  if (passphrase) {
    if (FAILED(hr = write(hWrite, "Passphrase: "))) goto finish_func;
    if (FAILED(hr = write(hWrite, passphrase))) goto finish_func;
    if (FAILED(hr = write(hWrite, "\r\n"))) goto finish_func;
  }
  if (FAILED(hr = write(hWrite, "%secring "))) goto finish_func;
  if (FAILED(hr = write(hWrite, secring))) goto finish_func;
  if (FAILED(hr = write(hWrite, "\r\n"))) goto finish_func;
  if (FAILED(hr = write(hWrite, "%pubring "))) goto finish_func;
  if (FAILED(hr = write(hWrite, pubring))) goto finish_func;
  if (FAILED(hr = write(hWrite, "\r\n"))) goto finish_func;
  if (FAILED(hr = write(hWrite, "%commit\r\n"))) goto finish_func;
  ::CloseHandle(hWrite);
  hWrite = INVALID_HANDLE_VALUE;
finish_func:
  return (hr);
}

HRESULT gpg::listkey(BOOL bSec, std::vector<std::string> &keylist, LPCSTR name, LPCSTR passphrase)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  std::string commandline;
  HANDLE hRead = INVALID_HANDLE_VALUE;
  commandline += " --fixed-list-mode --with-colons --with-fingerprint";
  if (bSec) {
    commandline += " --list-secret-keys";
  } else {
    commandline += " --list-keys";
  }
  if (name) {
    commandline += name;
  }
  if (FAILED(hr = createReadPipe(&hRead, &hStdOut))) {
    goto finish_func;
  }
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;

  {
    std::string str, str3, strid, strfpr;
    DWORD dwRead;
    keylist.clear();
    while (SUCCEEDED(hr = read(hRead, str, &dwRead))) {
      if (dwRead == 0) break;
      if ((str3 == "pub") || (str3 == "sec")) {
        if (str.substr(0, 3) == "uid") {
          int ipos = 0;
          for (int i = 0; i < 9; ++i) {
            ipos = str.find(":", ipos + 1);
          }
          ++ipos;
          strid = str.substr(ipos, str.find(":", ipos) - ipos);
        } else if (str.substr(0, 3) == "fpr") {
          int ipos = 0;
          for (int i = 0; i < 9; ++i) {
            ipos = str.find(":", ipos + 1);
          }
          ++ipos;
          strfpr = str.substr(ipos, str.find(":", ipos) - ipos);
        } else {
          if (!strfpr.empty()) {
            keylist.push_back(strid + ":" + strfpr);
            strid = "";
            strfpr = "";
          }
          str3 = str.substr(0, 3);
        }
      } else {
        if (!strfpr.empty()) {
          keylist.push_back(strid + ":" + strfpr);
          strid = "";
          strfpr = "";
        }
        str3 = str.substr(0, 3);
      }
    }
    if (!strfpr.empty()) {
      keylist.push_back(strid + ":" + strfpr);
    }
  }

  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
finish_func:
  return (hr);
}

HRESULT gpg::importkey(LPCSTR keyfile, LPCSTR passphrase)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  std::string commandline;

  commandline += " --batch --yes";
  commandline += " --import";
  commandline += " \"";
  commandline += keyfile;
  commandline += "\"";
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (dwExitCode == 0) {
    hr = S_OK;
  } else {
    hr = E_FAIL;
  }
finish_func:
  return (hr);
}

HRESULT gpg::exportkey(LPCSTR keyid, LPCSTR keyfile, LPCSTR passphrase)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  std::string commandline;

  commandline += " --batch --yes";
  commandline += " --export";
  commandline += " -o \"";
  commandline += keyfile;
  commandline += "\"";
  commandline += " \"";
  commandline += keyid;
  commandline += "\"";
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (dwExitCode == 0) {
    hr = S_OK;
  } else {
    hr = E_FAIL;
  }
finish_func:
  return (hr);
}

HRESULT gpg::deletekey(LPCSTR keyid, LPCSTR passphrase)
{
  HRESULT hr = S_OK;
  DWORD dwExitCode = 0;
  std::string commandline;

  commandline += " --batch --yes";
  commandline += " --delete-key";
  commandline += " \"";
  commandline += keyid;
  commandline += "\"";
  if (FAILED(hr = execute(commandline, passphrase))) goto finish_func;
  if (FAILED(hr = close(&dwExitCode))) goto finish_func;
  if (dwExitCode == 0) {
    hr = S_OK;
  } else {
    hr = E_FAIL;
  }
finish_func:
  return (hr);
}
